!******************************************************************************!
!                                                                              !
!    random_psp_rcl.f90                                                        !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2020 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!

module random_psp_rcl

!------------------------------------------------------------------------------!

  implicit none

!------------------------------------------------------------------------------!

  contains

!------------------------------------------------------------------------------!

  subroutine set_outgoing_momenta_rcl(npr,pIn,p)

    use globals_rcl
#ifdef WITHRAMBO
    use aafurambo, only: rambo
#endif

    ! This subroutine sets the momenta for the process with process
    ! number "npr". "pIn" are the incoming momenta.
    ! "p" are all momenta (incoming and outgoing) as generated by the
    ! subroutine (of course the incoming momenta coincide with "pIn").
    ! The subroutine uses RAMBO to generate the momenta.

    integer,  intent(in)  :: npr
    real(dp), intent(in)  :: pIn(0:,:)
    real(dp), intent(out) :: p(0:,:)

#ifdef WITHRAMBO
    integer               :: pr,i,ip,lin,lout,l
    real(dp), allocatable :: pOutR(:,:),m(:)
    real(dp)              :: roots,wt,pCM(0:3),beta(1:3),beta2,gamma, &
                             E,pv(1:3),betap

    call get_pr(npr, 'set_outgoing_momenta_rcl', pr)
    call processes_not_generated_error_rcl('set_outgoing_momenta_rcl')

    lin  = prs(pr)%legsIn
    lout = prs(pr)%legsOut
    l    = lin + lout

    do i = 0,3
      pCM(i) = sum(pIn(i,:))
    enddo
    roots = sqrt(pCM(0)**2-sum(pCM(1:3)**2))
    beta(1:3) = pCM(1:3)/pCM(0)
    beta2 = sum(beta(1:3)**2)
    gamma = pCM(0)/roots

    allocate (m(l))

    do i = 1,l
      if (prs(pr)%crosspr .gt. 0) then
        ip = prs(pr)%relperm(i)
        m(i) = mONS(newleg(ip,prs(pr)%crosspr),prs(pr)%crosspr)
      else
        m(i) = mONS(newleg(i,pr),pr)
      end if
    enddo

    allocate (pOutR(4,lout))

    call rambo(lout,roots,m(lin+1:l),pOutR,wt,0)

    if (beta2.eq.0d0) then
      p(0,lin+1:l) = pOutR(4,1:lout)
      p(1,lin+1:l) = pOutR(1,1:lout)
      p(2,lin+1:l) = pOutR(2,1:lout)
      p(3,lin+1:l) = pOutR(3,1:lout)
    else
      do i = 1,lout
        E = pOutR(4,i)
        pv(1:3) = pOutR(1:3,i)
        betap = sum(beta(:)*pv(:))
        p(0,lin+i) = gamma*( E + betap )
        p(1:3,lin+i) = &
        pv(1:3) + beta(1:3)*( gamma*E - betap/beta2*(1d0-gamma) )
      enddo
    endif

    deallocate (pOutR)

    deallocate (m)
#else
    call error_rcl('Recola2 not compiled with RAMBO.', &
                   where='set_outgoing_momenta_rcl')
#endif

  end subroutine set_outgoing_momenta_rcl

!------------------------------------------------------------------------------!

end module random_psp_rcl

!------------------------------------------------------------------------------!
